#include <stdatomic.h>
#include <pthread.h>

/* Some weird windows compilers (i. e. not msvc) actually
 * support POSIX threads, but this two includes, usually,
 * are Unix specific. */
#include <sys/types.h>
#include <unistd.h>

#include <syscall.h> /* Linux specific */

#include <stdlib.h>
#include <stdio.h>

#define THREAD_NR	16

struct tas_lock {
	atomic_flag locked;
};

static void tas_lock_init(struct tas_lock *lock)
{
	atomic_flag_clear(&lock->locked);
}

static void tas_lock(struct tas_lock *lock)
{
	while (atomic_flag_test_and_set(&lock->locked));
}

static void tas_unlock(struct tas_lock *lock)
{
	atomic_flag_clear(&lock->locked);
}

static atomic_int cont;

static int thread_id(void)
{
	/* linux specific */
	return syscall(SYS_gettid);
}

static void *worker(void *data)
{
	struct tas_lock *lock = data;
	const int id = thread_id();

	while (atomic_load(&cont)) {
		tas_lock(lock);
		printf("Thread %d acquired lock\n", id);
		sleep(1); /* declared in unistd.h */
		printf("Thread %d released lock\n", id);
		tas_unlock(lock);
	}

	return 0;
}

int main(void)
{
	pthread_t ths[THREAD_NR];
	struct tas_lock lock;

	tas_lock_init(&lock);
	atomic_store(&cont, 1);

	for (int i = 0; i != THREAD_NR; ++i) {
		const int rc = pthread_create(&ths[i], NULL, &worker, &lock);

		if (rc) {
			perror("pthread_create");
			exit(2);
		}
	}

	sleep(5); /* declared in unistd.h */
	atomic_store(&cont, 0);

	for (int i = 0; i != THREAD_NR; ++i)
		pthread_join(ths[i], NULL);

	return 0;
}
